import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import datasets
from torchvision import transforms
from DashAI.back.models.base_model import BaseModel
from DashAI.back.config_object import ConfigObject
from DashAI.back.core.schema_fields import BaseSchema, schema_field, int_field, float_field, list_field

class MLPImageClassifierSchema(BaseSchema):
    epochs: schema_field(
        int_field(ge=1),
        placeholder=3,
        description="Number of training epochs"
    )
    learning_rate: schema_field(
        float_field(gt=0, lt=1),
        placeholder=0.001,
        description="Learning rate for the optimizer"
    )
    hidden_dims: schema_field(
        list_field(item_type=int_field(gt=0)),
        placeholder=[128, 128, 128],
        description="List of hidden layer dimensions"
    )

class MLP(nn.Module):
    def __init__(self, input_dim, output_dim, hidden_dims):
        super().__init__()
        self.hidden_layers = nn.ModuleList()
        previous_dim = input_dim

        for hidden_dim in hidden_dims:
            self.hidden_layers.append(nn.Linear(previous_dim, hidden_dim))
            previous_dim = hidden_dim

        self.output_layer = nn.Linear(previous_dim, output_dim)
        self.relu = nn.ReLU()

    def forward(self, input: torch.Tensor):
        batch_size = input.shape[0]
        x = input.view(batch_size, -1)

        for layer in self.hidden_layers:
            x = self.relu(layer(x))

        x = self.output_layer(x)
        return x

class TransformedImageDataset(torch.utils.data.Dataset):
    def __init__(self, dataset: datasets.Dataset):
        self.dataset = dataset
        self.transforms = transforms.Compose(
            [
                transforms.Resize((30, 30)),
                transforms.ToTensor(),
            ]
        )

        column_names = list(self.dataset.features.keys())
        self.image_col_name = column_names[0]
        self.tensor_shape = self.transforms(self.dataset[0][self.image_col_name]).shape
        self.input_dim = self.tensor_shape[0] * self.tensor_shape[1] * self.tensor_shape[2]
        if len(column_names) > 1:
            self.label_col_name = column_names[1]
            self.output_dim = len(set(self.dataset[self.label_col_name]))
        else:
            self.label_col_name = None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        image = self.dataset[idx][self.image_col_name]
        image = self.transforms(image)
        if self.label_col_name is None:
            return image
        label = self.dataset[idx][self.label_col_name]
        return image, label

class MLPImageClassifier(BaseModel, ConfigObject):
    SCHEMA = MLPImageClassifierSchema
    COMPATIBLE_COMPONENTS = ["ImageClassificationTask"]

    def __init__(self, epochs, learning_rate, hidden_dims, **kwargs):
        super().__init__(**kwargs)
        self.epochs = epochs
        self.hidden_dims = hidden_dims
        self.learning_rate = learning_rate
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = None

    def fit(self, x_train: datasets.Dataset, y_train: datasets.Dataset):
        dataset = datasets.Dataset.from_dict(
            {
                "image": x_train["image"],
                "label": y_train["label"],
            }
        )
        image_dataset = TransformedImageDataset(dataset)
        train_loader = DataLoader(image_dataset, batch_size=32, shuffle=True)

        self.model = MLP(image_dataset.input_dim, image_dataset.output_dim, self.hidden_dims).to(self.device)
        criterion = nn.CrossEntropyLoss()
        optimizer = optim.Adam(self.model.parameters(), lr=self.learning_rate)

        self.model.train()
        for epoch in range(self.epochs):
            for images, labels in train_loader:
                images, labels = images.to(self.device), labels.to(self.device)
                optimizer.zero_grad()
                outputs = self.model(images)
                loss = criterion(outputs, labels)
                loss.backward()
                optimizer.step()
            print(f"Epoch {epoch+1}/{self.epochs}, Loss: {loss.item()}")

    def predict(self, x_pred: datasets.Dataset):
        image_dataset = TransformedImageDataset(x_pred)
        test_loader = DataLoader(image_dataset, batch_size=32, shuffle=False)
        
        self.model.eval()
        probs_predicted = []
        with torch.no_grad():
            for images in test_loader:
                images = images.to(self.device)
                output_probs: torch.Tensor = self.model(images)
                probs_predicted += output_probs.tolist()
        return probs_predicted

    def save(self, filename: str) -> None:
        torch.save(self.model, filename)

    @staticmethod
    def load(filename: str):
        return torch.load(filename)
